package npac.webmail;

import java.io.*;
import java.util.*;

/**
 * This class parses a message folder.  The filename can be relative or 
 * absolute.
 *
 * This will allow a user to give a filename and read the contents of
 * the file in mail format.  The contents of the file are returned in special
 * structures.
 *
 * For example:
 * <pre>
 *    FolderParser fp = new FolderParser("foobar","machine1.some.univ.edu","/user/foobar/folders/folder1");
 *    IndexFormat index = fp.index ();
 *    MailFormat mailf = fp.body (i);
 * </pre>
 *
 * @see            webmail.imap.MailFormat
 * @see            webmail.imap.IndexFormat
 * @see            webmail.imap.AddressFormat
 * @version        1.0 21 October 1996
 * @author         Mohan Gummalam
 */
public class FolderParser extends FolderSuperclass {
  private File file;
  private String user;
  private String machineName;

  private IndexFormat iformat;
  private MailFormat[] mformat;

  private Vector vMails = new Vector ();
  private Vector vBody = new Vector ();

  private int numTotalMails = 0;
  private int numNewMails = 0;
  private int numAnsweredMails = 0;
  private int numDeletedMails = 0;

  private Vector vIndFrom = new Vector ();
  private Vector vIndDate = new Vector ();
  private Vector vIndSubject = new Vector ();
  private Vector vIndMsgSize = new Vector ();
  private Vector vIndIsA = new Vector ();
  private Vector vIndIsF = new Vector ();
  private Vector vIndIsD = new Vector ();
  private Vector vIndIsS = new Vector ();

  private Vector vTemp = new Vector ();

  private static final int 
	      NONE = 0,
              DATE = 1, FROM = 2, TO = 3, CC = 4, BCC = 5, SUBJECT = 6,
              STATUS = 7, XSTATUS = 8,
              MSGID = 9, MIMEVER = 10, CONTENTTYPE = 11, // Not Implemented
              SENDER = 12, REPLYTO = 13,
              HEADER = 100, BODY = 200;

  public FolderParser ( String u, String m, File f ) {
    user = u;
    machineName = m;
    file = f;
    selectMailbox ( file.getName() );
  }

  public FolderParser ( String u, String m, String fname ) {
    user = u;
    machineName = m;
    file = new File ( fname );
    selectMailbox ( file.getName() );
  }

  public FolderParser ( String u, String m, String dname, String fname ) {
    user = u;
    machineName = m;
    file = new File ( dname, fname );
    selectMailbox ( file.getName() );
  }

/**
 * Parses the file, and (re)constructs the index.  An explicit call to this
 * function has to be made, before making a call to <tt>getIndex</tt> or
 * <tt>getMail</tt>.
 */
  public int selectMailbox ( String fname ) {
    int state = NONE, substate = NONE;
    MailFormat mf = null;
    AddressFormat addr = null;
    int i=-1;
    int messageSize = 0;
    String line = null;
    RandomAccessFile raf = null;
    if (!file.exists()) {
      System.out.println ("File doesnt exist! Exiting");
//      throw ( new NoSuchFolderException() );
    }
    try {
      raf = new RandomAccessFile ( file, "rw" );
    } catch (IOException ioe) {
      System.out.println ( "Failed to open file: " + file +"\nExiting.\n");
//      throw ( ioe );
    }
    try {
      while ( (line = raf.readLine()) != null ) {
	if ( mailHeader( line ) ) {
	  i++;
System.out.println ( "Message ["+i+"] " + line );
	  state = HEADER;
	  vIndFrom.addElement ( "" );
	  vIndDate.addElement ( "" );
	  vIndSubject.addElement ( "" );
	  vIndMsgSize.addElement ( "" );
	  numNewMails++;
	  vIndIsA.addElement ( new Boolean( false ) );
	  vIndIsF.addElement ( new Boolean( false ) );
	  vIndIsD.addElement ( new Boolean( false ) );
	  vIndIsS.addElement ( new Boolean( false ) );
	  //  since this marks the beginning of new mail, save the contents of
	  //  previous mail.  Process the new mail after this point.
	  //  Take care of the first time..
	  if ( mf != null ) {	//  if (not the first time)
	    mf.setMessageBody (vBody);
            mf.setMessageSize ( ""+messageSize );
	    vIndMsgSize.setElementAt ( ""+messageSize, i-1 );
            messageSize = 0;
	    vMails.addElement (mf);
	    mf = null;
	    vBody.removeAllElements ();
	  }
	  mf = new MailFormat ();
	}
	else {
	  messageSize += line.length()+2;	// +2 for 'CR'+'LF'
	  if ( state == BODY )
	    vBody.addElement ( line );
	  else if ( line.startsWith("Date: ") && (state==HEADER) )  {
	    String ss = line.substring("Date: ".length());
	    vIndDate.setElementAt ( ss, i );
	    mf.setDate (ss);
	  }
	  else if ( line.startsWith("From: ") && (state==HEADER) ) {
	    substate = FROM;
	    ParseAddressString pas = new ParseAddressString ( line.substring("From: ".length()) );
	    AddressFormat af = pas.parse();
	    vIndFrom.setElementAt ( af, i );
	    mf.setFrom (af);
	    mf.setSender (af);
	    mf.setReplyTo (af);
	  }
	  else if ( line.startsWith("Sender: ") && (state==HEADER) ) {
	    substate = SENDER;
	    ParseAddressString pas = new ParseAddressString ( line.substring("Sender: ".length()) );
	    AddressFormat af = pas.parse();
	    mf.setSender (af);
	  }
	  else if ( line.startsWith("Reply-To: ") && (state==HEADER) ) {
	    substate = REPLYTO;
	    ParseAddressString pas = new ParseAddressString ( line.substring("Reply-To: ".length()) );
	    AddressFormat af = pas.parse();
	    mf.setReplyTo (af);
	  }
	  else if ( line.startsWith("To: ") && (state==HEADER) ) {
	    substate = TO;
	    // Since this is the start of a new state, remove any elements that
	    // _might_ have been filled into this vector before using it.
	    vTemp.removeAllElements ();
	    ParseAddressString pas = new ParseAddressString ( line.substring("To: ".length()) );
	    AddressFormat af = null;
	    while ( (af = pas.parse()) != null )
	      vTemp.addElement ( af );
	    mf.setTo ( vTemp );
	  }
	  else if ( (line.startsWith("Cc: ") || line.startsWith("cc: ")) && (state==HEADER) ) {
	    substate = CC;
	    // Since this is the start of a new state, remove any elements that
	    // _might_ have been filled into this vector before using it.
	    vTemp.removeAllElements ();
		ParseAddressString pas=null;  
		if(line.startsWith("Cc: "))
	    	pas = new ParseAddressString ( line.substring("Cc: ".length()) );
		else if(line.startsWith("cc: "))
	    	pas = new ParseAddressString ( line.substring("cc: ".length()) );
	    AddressFormat af = null;
	    while ( (af = pas.parse()) != null )
	      vTemp.addElement ( af );
	    mf.setCc ( vTemp );
	  }
	  else if ( line.startsWith("Bcc: ") && (state==HEADER) ) {
	    substate = BCC;
	    // Since this is the start of a new state, remove any elements that
	    // _might_ have been filled into this vector before using it.
	    vTemp.removeAllElements ();
	    ParseAddressString pas = new ParseAddressString ( line.substring("Bcc: ".length()) );
	    AddressFormat af = null;
	    while ( (af = pas.parse()) != null )
	      vTemp.addElement ( af );
	    mf.setBcc ( vTemp );
	  }
	  else if ( line.startsWith("Subject: ") && (state==HEADER) ) {
	    substate = SUBJECT;
	    String ss = line.substring("Subject: ".length());
	    vIndSubject.setElementAt ( ss, i );
	    mf.setSubject (ss);
	  }
	  else if ( line.startsWith("Status: ") && (state==HEADER) ) {
	    substate = STATUS;
	    String str = line.substring("Status: ".length());
	    mf.setIsSeen ( false );
	    if (str.indexOf('R') != -1) {
	      numNewMails--;
	      vIndIsS.setElementAt ( new Boolean( true ), i );
	      mf.setIsSeen ( true );
	    }
	  }
	  else if ( line.startsWith("X-Status: ") && (state==HEADER) ) {
	    substate = XSTATUS;
	    String str = line.substring("X-Status: ".length());
	    mf.setIsAnswered ( false );
	    if (str.indexOf('A') != -1) {
	      numAnsweredMails++;
	      vIndIsA.setElementAt ( new Boolean( true ), i );
	      mf.setIsAnswered ( true );
	    }
	    
	    mf.setIsDeleted ( false );
	    if (str.indexOf('D') != -1) {
	      numDeletedMails++;
	      vIndIsD.setElementAt ( new Boolean( true ), i );
	      mf.setIsDeleted ( true );
	    }
	  }
	  else if ( line.equals("") && (state==HEADER) ) {
	    state = BODY;
	    substate = NONE;
	  }
	  // Case when the To: (and similarly for Cc: and Bcc:.  The Bcc: line
	  // will appear in a send-mail folder only) list exceeds one line.  So
	  // here the 'spill over' lines are being considered!
	  else if ( substate == TO ) {
	    ParseAddressString pas = new ParseAddressString ( line );
	    AddressFormat af = null;
	    while ( (af = pas.parse()) != null )
	      vTemp.addElement ( af );
	    // this setTo overrides the previous setTo we had.  Since the Vector
	    // must not have changed since the last setTo, we have preserved the
	    // To: list.
	    mf.setTo ( vTemp );
	  }
	  else if ( substate == CC ) {
	    ParseAddressString pas = new ParseAddressString ( line );
	    AddressFormat af = null;
	    while ( (af = pas.parse()) != null )
	      vTemp.addElement ( af );
	    mf.setCc ( vTemp );
	  }
	  else if ( substate == BCC ) {
	    ParseAddressString pas = new ParseAddressString ( line );
	    AddressFormat af = null;
	    while ( (af = pas.parse()) != null )
	      vTemp.addElement ( af );
	    mf.setBcc ( vTemp );
	  }
	}
      }
      // For the last mail!  Storing it into the mailFormat array.
      mf.setMessageBody (vBody);
      mf.setMessageSize ( ""+messageSize );
      vIndMsgSize.setElementAt ( ""+messageSize, i );
      vMails.addElement (mf);
    } catch (IOException ioe2) {
      System.out.println ( "Failed to open file: " + file +"\nExiting.\n");
//      throw ( ioe2 );
    }
    numTotalMails = i+1;
    iformat = new IndexFormat ( numTotalMails );
    iformat.setFrom ( vIndFrom );
    iformat.setDate ( vIndDate );
    iformat.setSubject ( vIndSubject );
    iformat.setMessageSize ( vIndMsgSize );

    iformat.setIsDeleted ( vIndIsD );
    iformat.setIsSeen ( vIndIsS );
    iformat.setIsFlagged ( vIndIsF );
    iformat.setNumUnread ( numNewMails );
    iformat.setNumAnswered ( numAnsweredMails );
    iformat.setNumNew ( numNewMails );
    iformat.setNumDeleted ( numDeletedMails );
    mformat = new MailFormat [numTotalMails];
    vMails.copyInto ( mformat ); 
    for(int j=0; j<mformat.length; j++)
      System.out.println("Mailformat :"+j+"\n**********************************************************************\n"+mformat[j]);
    return numTotalMails;
  }

/**
 * Return the list of mails in the file.  The returned value is of the
 * <tt>IndexFormat</tt> class.
 * @return           The index.
 */
  public IndexFormat index () { return iformat; }

/**
 * Return a specified mail.  The returned value is of the <tt>MailFormat</tt>
 * class.
 * @return           The index.
 */
  public MailFormat body ( int c ) {
    mformat[c-1].setIsSeen(true);
    iformat.setIsSeen(c-1,true);
    return mformat[c-1];
  }



/**
 * Save into the file a specified mail number. 
 * @return           <tt>true</tt> if save is successful, <tt>false</tt> otherwise
 */
  public void save ( int c, File folder ) throws NoSuchFolderException {
    if (!folder.exists())
      throw ( new NoSuchFolderException () );
    saveIntoFile(c,folder);
  }


/**
 * Save into the file a specified mail number. 
 * @return           <tt>true</tt> if save is successful, <tt>false</tt> otherwise
 */
  public void save ( int c, String fullFilePath) throws NoSuchFolderException {
    File f = new File (fullFilePath); 
    if (!f.exists())
      throw ( new NoSuchFolderException () );
    saveIntoFile (c, f);
  }


/**
 * Save into the file a specified mail number. 
 * @return           <tt>true</tt> if save is successful, <tt>false</tt> otherwise
 */
  public void save ( int c, String directory, String filename ) throws NoSuchFolderException {
    File f = new File (directory, filename);
    if (!f.exists())
      throw ( new NoSuchFolderException () );
    saveIntoFile (c, f);
  }

  private void saveIntoFile ( int c, File file ) throws NoSuchFolderException {
    RandomAccessFile localRAF = null;
    try {
      localRAF = new RandomAccessFile ( file, "rw" );
      localRAF.seek ( localRAF.length() );
      Date d = new Date();
      mformat[c-1].setIsDeleted( false );
      String x = "From "+user+"@"+machineName+" "+d.toLocaleString()+"\n"+mformat[c-1];
      localRAF.writeBytes ( x );
      localRAF.close();
    } catch (IOException ioe) {
      throw ( new NoSuchFolderException () );
    }
  }

/**
 * To mark the mail 'deleted'/'undeleted'.
 * @return           <tt>true</tt> if delete was successful, <tt>false</tt> otherwise
 */
  public boolean delete ( int c, boolean dele ) { 
    iformat.setIsDeleted(c-1,dele);
    mformat[c].setIsDeleted( dele );
    return true;
  }

/**
 * Expunge the mail from this folder.
 * @return           <tt>true</tt> if delete was successful, <tt>false</tt> otherwise
 */
  public void expunge () throws IOException {
    Vector v = new Vector();
    for (int i=0; i<numTotalMails; i++)
      v.addElement(mformat[i]);
    for (int i=0; i<numTotalMails; i++) {
      MailFormat mf = (MailFormat)v.elementAt(i);
      if ( mf.getIsDeleted() ) {
	if (mf.getIsAnswered())
	  numAnsweredMails--;
	if (!mf.getIsSeen())
	  numNewMails--;
	if (mf.getIsDeleted())
	  numDeletedMails--;
	v.removeElementAt(i);
	numTotalMails--;
      }
    }
    mformat = null;
    mformat = new MailFormat [v.size()];
    v.copyInto (mformat);
    RandomAccessFile localRAF = null;
    localRAF = new RandomAccessFile ( file, "rw" );
    Date d = new Date();
    for (int j=0; j<numTotalMails; j++) {
      String x = "From "+user+"@"+machineName+" "+d+"\n"+mformat[j-1];
      localRAF.writeBytes ( x );
    }
    localRAF.close();
  }

/**
 * Overrides the <tt>Object.toString()</tt> function.
 * @return           the object converted to a string.
 */
  public String toString () {
    String outString = "";

    outString += "[Folder: " + file + "]\n";
    outString += "[        numTotalMails: " + numTotalMails + "]\n";
    outString += "[        numNewMails: " + numNewMails + "]\n";
    outString += "[        numAnsweredMails: " + numAnsweredMails + "]\n";
    outString += "[        numDeletedMails: " + numDeletedMails + "]\n";
    outString += "---------------------------------------------------------\n";
    for (int i=0; i<numTotalMails; i++)
      outString += "[" + (((Boolean)vIndIsA.elementAt(i)).booleanValue() ? "A" : "_") +  (((Boolean)vIndIsD.elementAt(i)).booleanValue() ? "D" : "_") + (((Boolean)vIndIsS.elementAt(i)).booleanValue() ? "_" : "N") + "] [" + (i+1) + "]  [" + vIndDate.elementAt(i) + "]  [" + ((AddressFormat)vIndFrom.elementAt(i)) + "]  [" + vIndSubject.elementAt(i) + "]\n";
    return outString;
  }

  // This is slightly hard-coding of the parsing of the mail header.
  private boolean mailHeader ( String s ) {
    if ( !s.startsWith("From ") )
      return false;
    String ss = s.substring("From ".length());
    int indexOfDate = ss.indexOf(' ');
    if (ss.substring(0, indexOfDate).indexOf('@') == -1)
      return false;	// No '@' found..  so not an email address.
    ss = ss.substring(indexOfDate).trim();
    if (!ss.startsWith("Sun") && !ss.startsWith("Mon") &&
	!ss.startsWith("Tue") && !ss.startsWith("Wed") &&
	!ss.startsWith("Thu") && !ss.startsWith("Fri") &&
	!ss.startsWith("Sat"))
      return false;
    ss = ss.substring(4).trim();
    if (!ss.startsWith("Jan") && !ss.startsWith("Feb") &&
	!ss.startsWith("Mar") && !ss.startsWith("Apr") &&
	!ss.startsWith("May") && !ss.startsWith("Jun") &&
	!ss.startsWith("Jul") && !ss.startsWith("Aug") &&
	!ss.startsWith("Sep") && !ss.startsWith("Oct") &&
	!ss.startsWith("Nov") && !ss.startsWith("Dec"))
      return false;
    ss = ss.substring(4).trim();
    try {
      int date = Integer.parseInt(ss.substring(0, 3).trim());
      if (date>31 || date<1)
	return false;
    }
    catch (NumberFormatException nfe) {
      return false;
    }
    ss = ss.substring(3).trim();
    // premature termination of the parsing.
    // Have to parse time and timezone, and year.
    // For the time being I think this should suffice.
    // Mohan. 10/27/96

    return true;
  }

  public static void main ( String argv[] ) {
    FolderParser folp = new FolderParser ( "furma", "postoffice.npac.syr.edu", "/home/B8F/furma/mail", argv[0] );
    System.out.println(folp);
  }

}

// This class assumes address string to be in the following form.
// <address string> ::= <address> | <address> "," <whitespace> <address string>
// <address> ::= "\"" <personal name> "\"" <whitespace> "<" <mailboxname> "@" <domain name> ">" |
//		<personal name> <whitespace> "<" <mailboxname> "@" <domain name> ">" |
//		<mailboxname> "@" <domain name>
//  Minimal error checking is performed.
class ParseAddressString {
  String baseString = null;
  String tempString = null;
  int current, previous;
  int length;

  public ParseAddressString ( String s ) {
    baseString = s;
    tempString = s;
    current = previous = 0;
    length = s.length();
  }

  public AddressFormat parse () {
    while ( (current < length) &&
	    ((baseString.charAt(current) == ' ') ||
	     (baseString.charAt(current) == ',') ||
	     (baseString.charAt(current) == '\t'))
	   )
      current++;
    if (current >= length)
      return null;
    if ( baseString.charAt(current) == '"' ) {
      previous = ++current;
      boolean escape = false;
      while (current<length && (baseString.charAt(current) != '"') && !escape) {
	if (baseString.charAt(current) == '\\') {
	  if (!escape)
	    escape = true;
	  else if (escape)
	    escape = false;
	}
	else
	  if (escape)
	    escape = false;
	current++;
      }
      // premature ending of the address.  with just the personal name!
      if (current>=length) return null;
      String personalName = baseString.substring (previous, current);
      if (current<length && baseString.charAt(current) != '<')
	current++;
      // premature ending of the address.  with just the personal name!
      if (current>=length) return null;
      previous = ++current;
      int i = baseString.indexOf('@', previous);
      current = baseString.indexOf ('>', previous);
      // premature ending of the address.  doesnt contain an '@' or '>'
      if (i==-1 || current==-1) return null;
      return ( new AddressFormat ( personalName,
				   "",
				   baseString.substring(previous, i),
				   baseString.substring(i+1, current)));
    }
    else {
      previous = current;
      while ( current<length && 
	      (baseString.charAt(current) != '@') &&
	      (baseString.charAt(current) != ' ')
	     )
	current++;
      if (current>=length) {
	System.out.println ("***** ERROR: OUT OF BOUNDS!");
	return null;
      }
      if (baseString.charAt(current) == ' ') {	// The string is a personal
	current = baseString.indexOf('<', previous); // name, but not in "'s.
	String personalName = baseString.substring (previous, current).trim ();
	previous = ++current;
	current = baseString.indexOf ('>', previous);
	int i = baseString.indexOf ('@', previous);
	current++;
	// premature ending of the address.  doesnt contain an '@' or '>'
	if (i==-1 || current==-1) return null;
	return ( new AddressFormat ( personalName,
				     "",
				     baseString.substring(previous, i),
				     baseString.substring(i+1, current-1)));
      }
      else if (baseString.charAt(current) == '@') {	// The string is an
	int i = current;
	previous = ++current;
	while (current<length && (baseString.charAt(current) != ','))
	  current++;
	if (current>=length) return null;
	return ( new AddressFormat ( "",
				     "",
				     baseString.substring(previous, i),
				     baseString.substring(i+1, current)));
      }
      else
	return null;
    }
  }

  //  Currently an incomplete function.  Will have to complete this on a need
  //  basis.
  public AddressFormat parseReturnString () {
    return null;
  }
}
